#include "def.h"
#include "dbx.h"
#include "dbxexception.h"
#include <string>
#include <fstream>
#include <sstream>

void dbx::readmessage(uint5 pos)
{
  verify(pos);

  uint5 len=get5(pos+4);
  uint5 itemcount=get5(pos+8);

  int count=static_cast<int>((itemcount&0xff0000)>>16);

  if (count==0)
    throw dbxexception("Deleted message");

  message m;
  m.set_int("X-DBXTool-Itemcount-high-byte",itemcount>>24);
  m.set_int("X-DBXTool-Itemcount-low-word",itemcount&0xffff);

  // Read the items table
  for (int i=0;i<0x1d;++i)
    gotf_[i]=false;
  msgpos_=pos+12;
  for (int i=0;i<count;++i,msgpos_+=4)
    setitem(get5(msgpos_));

  if (!gotf_[0])
    throw dbxexception("No index!");
  m.set_int("X-DBXTool-Index",intfromitem(0));

  // Process the flags
  if (!gotf_[1])
    throw dbxexception("No flags!");
  uint5 flags=intfromitem(1);
  m.set_int("X-DBXTool-Flags",flags);
  if ((flags&0xf8250614)!=0)
    throw dbxexception("Unusual flags");

  bool gotbody=((flags&1)!=0);
  m.set_bool("X-DBXTool-Got-body",gotbody);
  if (gotbody!=gotf_[4])
    throw dbxexception("Flags and items disagree about body existence!");

  if ((flags&2)!=0)
    throw dbxexception("Message gone!");

  m.set_bool("X-DBXTool-Composing",(flags&8)!=0);
  m.set_bool("X-DBXTool-Flagged",(flags&0x20)!=0);

  if ((flags&0x40)!=0)
    throw dbxexception("Message broken!");

  m.set_bool("X-DBXTool-Read",(flags&0x80)!=0);
  m.set_bool("X-DBXTool-Marked-for-download",(flags&0x100)!=0);
  bool news=((flags&0x800)!=0);
  m.set_bool("X-DBXTool-News",news);
  m.set_bool("X-DBXTool-Signed",(flags&0x1000)!=0);
  m.set_bool("X-DBXTool-Encrypted",(flags&0x2000)!=0);
  m.set_bool("X-DBXTool-Attachment",(flags&0x4000)!=0);

  if ((flags&0x8000)!=0)
    throw dbxexception("Audio message!");

  m.set_bool("X-DBXTool-Flag11",(flags&0x20000)!=0);
  m.set_bool("X-DBXTool-Replied",(flags&0x80000)!=0);
  m.set_bool("X-DBXTool-Forwarded",(flags&0x100000)!=0);
  m.set_bool("X-DBXTool-Watch",(flags&0x400000)!=0);
  m.set_bool("X-DBXTool-Ignore",(flags&0x800000)!=0);
  m.set_bool("X-DBXTool-Flag18",(flags&0x1000000)!=0);

  if ((flags&0x2000000)!=0)
    throw dbxexception("Message hidden!");

  if ((flags&0x4000000)!=0)
    throw dbxexception("No received time flag set!");

  if (!gotf_[2])
    throw dbxexception("No received time!");
  m.set_int6("X-DBXTool-Received",int6fromitem(2));

  if (gotf_[3])
    m.set_int("X-DBXTool-Lines",intfromitem(3));

  if (gotf_[5])
    m.set_string("X-DBXTool-Conversation",stringfromitem(5));

  if (gotf_[6])
    m.set_int6("X-DBXTool-Obtained",int6fromitem(6));

  if (gotf_[7])
    m.set_string("X-DBXTool-Message-ID",stringfromitem(7));

  if (gotf_[8])
    m.set_string("X-DBXTool-Subject",stringfromitem(8));

  if (gotf_[9])
    m.set_string("X-DBXTool-Poster",stringfromitem(9));

  if (gotf_[0xa])
    m.set_string("X-DBXTool-References",stringfromitem(0xa));

  if (gotf_[0xb])
    m.set_string("X-DBXTool-XRef",stringfromitem(0xb));

  if (gotf_[0xc])
    m.set_string("X-DBXTool-Received-from",stringfromitem(0xc));

  if (gotf_[0xd])
    m.set_string("X-DBXTool-Sender-name",stringfromitem(0xd));

  if (gotf_[0xe])
    m.set_string("X-DBXTool-Sender-address",stringfromitem(0xe));

  if (gotf_[0xf])
    m.set_int("X-DBXTool-Item-0f",intfromitem(0xf));

  if (gotf_[0x10])
    m.set_int("X-DBXTool-X-Priority",intfromitem(0x10));
  else
    if (!news)
      throw dbxexception("No priority in email!");

  if (!gotf_[0x11])
    throw dbxexception("No message length");
  uint5 totallength=intfromitem(0x11);

  if (gotf_[0x12])
    m.set_int6("X-DBXTool-Sent",int6fromitem(0x12));

  if (gotf_[0x13])
    m.set_string("X-DBXTool-Receiver-name",stringfromitem(0x13));

  if (gotf_[0x14])
    m.set_string("X-DBXTool-Receiver-address",stringfromitem(0x14));

  if (gotf_[0x16])
    m.set_string("X-DBXTool-POP3-UIDL",stringfromitem(0x16));

  if (gotf_[0x18])
    m.set_string("X-DBXTool-Account",stringfromitem(0x1a));

  if (gotf_[0x1b])
    if (news)
      m.set_string("X-DBXTool-Newsgroup-ID",stringfromitem(0x1b));
    else
      m.set_string("X-DBXTool-POP3-Message-Number",stringfromitem(0x1b));

  uint5 msgpos=(gotbody ? intfromitem(4) : 0);
  std::string msg;
  while (msgpos!=0) {
    verify(msgpos);
    uint5 len=get5(msgpos+4);
    uint5 count=get5(msgpos+8);  // Number of bytes to include from this segment
    uint5 next=get5(msgpos+12);  // Pointer to next message segment

    if (len!=0x200)
      throw dbxexception("Unusual segment length");
    if (count>len)
      throw dbxexception("Segment too long!");
    if (count<len && next!=0)
      throw dbxexception("Curiously short segment");

    msg+=getstr(msgpos+16,msgpos+16+count);
    msgpos=next;
  }

  if (news)
    totallength-=5;
  if (totallength!=msg.length() && gotf_[4])
    m.set_int("X-DBXTool-Message-length-overestimate",static_cast<uint5>(totallength-msg.length()));

  m.set_body(msg);
  proc_->process(m);
  ++messagesread_;
}

// Parse mailbox tables
void dbx::readnode(uint5 pos)
{
  if (isused(pos))
    return;
  verify(pos);

  uint5 unknown1=get5(pos+4);
  uint5 leftnode=get5(pos+8);
  uint5 rightnode=get5(pos+12);
  uint5 unknown2=get5(pos+16);
  uint5 unknown3=get5(pos+20);

  readnode(leftnode);
  readnode(rightnode);

  for (uint5 p=pos+24;p<pos+0x27c;p+=12) {
    uint5 msg=get5(p);
    uint5 middlenode=get5(p+4);
    uint5 unknown4=get5(p+8);

    if (msg!=0)
      readmessage(msg);

    if (middlenode!=0)
      readnode(middlenode);
  }
}

dbx::dbx(const std::string& filename,processor* p)
  : proc_(p), messagesread_(0), buf_(0)
{
  {
    std::ifstream f;
    f.exceptions(std::ios_base::badbit | std::ios_base::failbit | std::ios_base::eofbit);
    f.open(filename.c_str(),std::ios_base::binary);
    f.seekg(0,std::ios::end);
    size_=f.tellg();
    f.seekg(0,std::ios::beg);
    buf_=new uint3[size_];
    f.read(reinterpret_cast<char*>(buf_),size_);
  }

  if (get5(0)!=0xfe12adcf)
    throw dbxexception("This does not appear to be an Outlook Express .dbx file!");

  uint5 guid1=get5(4),guid2=get5(8),guid3=get5(12),guid4=get5(16);

  // normal folders
  if (guid1==0x6F74FDC5 && guid2==0x11D1E366 && guid3==0xC0004E9A && guid4==0xd409a34f) {
    {
      uint5 v1=get5(20),v2=get5(24),v3=get5(28),v4=get5(32);
      std::stringstream str;
      str << v1 << "." << v2 << "." << v3 << "." << v4;
      version_=str.str();
    }
    storename_=getstring(0x24c1);
    boxname_=getstring(0x25c1);

    isused(0);
    readnode(get5(0x30));
    totalmessages_=get5(0xc4);
    if (totalmessages_!=messagesread_)
      throw dbxexception("Not all messages read");
    return;
  }

  // folders.dbx file
  if (guid1==0x6F74FDC6 && guid2==0x11D1E366 && guid3==0xC0004E9A && guid4==0xd409a34f)
    return;

  // offline.dbx file
  if (guid1==0x26fe9d30 && guid2==0x11d21a8f && guid3==0x6000bfaa && guid4==0xc474d497)
    return;

  // pop3uidl.dbx file
  if (guid1==0x6F74FDC7 && guid2==0x11D1E366 && guid3==0xC0004E9A && guid4==0xd409a34f)
    return;

  throw dbxexception("I don't understand this type of .dbx file!");
}
