#ifndef DBX_H
#define DBX_H

#include "def.h"
#include "dbxexception.h"
#include "processor.h"
#include <string>
#include <vector>
#include <set>

class dbx
{
public:
  dbx(const std::string& filename,processor* p);
  ~dbx() { if (buf_!=0) delete[] buf_; }

private:
  void readmessage(uint5 pos);
  void readnode(uint5 pos);

  processor* proc_;
  uint5 messagesread_,totalmessages_;
  std::string version_,storename_,boxname_;

  void verify(uint5 p)
  {
    if (p!=get5(p))
      throw dbxexception("Table not found!");
  }


  // Buffer
  uint3* buf_;
  uint5 size_;
  uint3 get3(uint5 p)
  {
    if (p>=size_)
      throw dbxexception("DBX file broken!");
    return buf_[p];
  }
  uint4 get4(uint5 p)
  {
    if (p>=size_-1)
      throw dbxexception("DBX file broken!");
    return *reinterpret_cast<uint4*>(buf_+p);
  }
  uint5 get5(uint5 p)
  {
    if (p>=size_-3)
      throw dbxexception("DBX file broken!");
    return *reinterpret_cast<uint5*>(buf_+p);
  }
  uint6 get6(uint5 p)
  {
    if (p>=size_-7)
      throw dbxexception("DBX file broken!");
    return *reinterpret_cast<uint6*>(buf_+p);
  }
  std::string getstr(uint5 start,uint5 end)
  {
    if (start>end || end>=size_)
      throw dbxexception("DBX file broken!");
    return std::string(buf_+start,buf_+end);
  }
  std::string getstring(uint5 pos)
  {
    int endp=pos;
    while (get3(endp)!=0)
      ++endp;
    return getstr(pos,endp);
  }


  // Used node list
  std::set<uint5> used_;
  bool isused(uint5 node)
  {
    std::set<uint5>::const_iterator i=used_.find(node);
    bool uf=(i!=used_.end());
    used_.insert(node);
    return uf;
  }


  // Message currently being processed
  int msgpos_;
  bool absf_[0x1d];
  bool gotf_[0x1d];
  uint5 value_[0x1d];
  void setitem(uint5 v)
  {
    int item=v&0x7f;
    if (item>0x1c || item==0x15 || item==0x17 || item==0x18 || item==0x19)
      throw dbxexception("Unknown item");
    if (gotf_[item])
      throw dbxexception("Duplicate item");

    gotf_[item]=true;
    absf_[item]=((v&0x80)==0x80);
    value_[item]=(v>>8);
  }
  uint5 intfromitem(int item) { return (absf_[item] ? value_[item] : get5(msgpos_+value_[item])); }
  int6 int6fromitem(int item)
  {
    if (absf_[item])
      throw dbxexception("Absolute int6");
    return static_cast<int6>(get6(msgpos_+value_[item]));
  }
  std::string stringfromitem(int item)
  {
    if (absf_[item])
      throw dbxexception("Absolute string");
    return getstring(msgpos_+value_[item]);
  }
};

#endif // DBX_H
