#include "def.h"
#include "dbxexception.h"
#include "messagesaver.h"
#include <fstream>
#include <sstream>
#include <string>
#include <set>

messagesaver::messagesaver(std::string path)
 : path_(path)
{
  if (path_[path_.size()-1]!='\\')
    path_+='\\';
}

void messagesaver::process(message& m)
{
  std::string body=m.get_body();

  if (body=="")
    return;

  std::ofstream out;
  out.exceptions(std::ios_base::badbit | std::ios_base::failbit | std::ios_base::eofbit);

  int i=0;
  std::string fn;
  if (m.got_string("Subject"))
    fn=m.get_string("Subject");
  else {
    if (m.got_string("X-DBXTool-Subject"))
      fn=m.get_string("X-DBXTool-Subject");
    else {
      fn="";
      i=1;
    }
  }

  static const std::string forbidden="?*:\"<>\\/|"; // Characters we can't have in filenames

  // Maks the subject suitable for a filename
  fn=fn.substr(0,128); // Truncate so we don't get overly long paths.
  for (std::string::size_type p=0;p<fn.size();++p) {
    char c=fn[p];
    if ((c&0x80)!=0 || c<0x20) // Get rid of MBCS and control characters - they can cause problems
      fn[p]='_';
    else
      for (std::string::size_type q=0;q<forbidden.size();++q)
        if (c==forbidden[q]) {
          fn[p]='_';
          break;
        }
  }
  fn=path_+fn;

  std::string outfn;
  do {
    outfn=fn;
    if (i>0) {
      std::stringstream str;
      str << "(" << i << ")";
      outfn+=str.str();
    }
    if (m.get_bool("X-DBXTool-News"))
      outfn+=".nws";
    else
      outfn+=".eml";
    ++i;
  } while(exists(outfn));

  {
    out.open(outfn.c_str(),std::ios::out);
    out << m.get_body();
    out.close();
  }

  HANDLE hf=CreateFileA(outfn.c_str(),GENERIC_WRITE,0,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
  if (hf!=INVALID_HANDLE_VALUE) {
    FILETIME ft;
    int6 t=m.get_int6("X-DBXTool-Received");
    ft.dwHighDateTime=static_cast<DWORD>(t>>32);
    ft.dwLowDateTime=static_cast<DWORD>(t&0xffffffff);
    SetFileTime(hf,&ft,&ft,&ft);
    CloseHandle(hf);
  }
  else {
    std::stringstream str;
    str << "Error reopening file " << fn;
    throw dbxexception(str.str());
  }
}

bool messagesaver::exists(std::string fn)
{
  std::set<std::string>::const_iterator i=used_.find(fn);
  if (i!=used_.end())
    return true;
  used_.insert(fn);
  HANDLE hf=CreateFileA(fn.c_str(),0,0,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
  if (hf!=INVALID_HANDLE_VALUE) {
    CloseHandle(hf);
    return true;
  }
  return false;
}
