#include "def.h"
#include "timestats.h"
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <iomanip>

std::string timestats::datefromtime64(int6 t)
{
  // Compute time portion

  int nanosecs=static_cast<int>(t%(10*1000*1000))*100;
  t/=10*1000*1000;
  int seconds=static_cast<int>(t%60);
  t/=60;
  int minutes=static_cast<int>(t%60);
  int t5=static_cast<int>(t/60);
  int hours=t5%24;
  t5/=24;


  // Compute date portion

  static const int daysin400years=(365*400)+97; // 97 leap years in 400 years
  int year=1601+(t5/daysin400years)*400;
  t5%=daysin400years;

  static const int daysin100years=(365*100)+24; // 24 leap years in most centuries
  int century=t5/daysin100years;
  bool leap=(century>=3);
  if (century==4) { year+=300;         t5=daysin100years; }
             else { year+=century*100; t5%=daysin100years; }

  static const int daysin4years=(365*4)+1; // 1 leap year in most 4 year periods
  int fouryear=t5/daysin4years;
  leap=(fouryear>=24 && !leap);
  if (fouryear==25) { year+=99;         t5=daysin4years; }
               else { year+=fouryear*4; t5%=daysin4years; }

  int yearof4=t5/365;
  leap=(yearof4>=3 && !leap);
  if (yearof4==4) { year+=3;       t5=365; }
             else { year+=yearof4; t5%=365; }

  int day,month;
  if (leap && t5==59) { day=29; month=2; }
  else {
    if (leap && t5>59)
      --t5;
    static const int accumdays[]={0,31,59,90,120,151,181,212,243,273,304,334,365};
    for (month=1;accumdays[month]<t5;++month);
    day=1+t5-accumdays[month-1];
  }

  std::stringstream str;
  str << std::setw(4) << std::setfill('0') << year << "/";
  str << std::setw(2) << std::setfill('0') << month << "/";
  str << std::setw(2) << std::setfill('0') << day << ": ";
  return str.str();
}

void timestats::output(std::string ofname)
{
  std::ofstream out;
  out.exceptions(std::ios_base::badbit | std::ios_base::failbit | std::ios_base::eofbit);
  out.open(ofname.c_str(),std::ios_base::out | std::ios_base::trunc);

  int6 period=10000000LL*60L*60L*24L*7L;
  int b0=static_cast<int>(mint_/period);
  int6 t0=b0*period;
  int nb=1+static_cast<int>(maxt_/period)-b0;
  std::vector<int> buckets;
  buckets.resize(nb);

  for (std::vector<int6>::const_iterator i=times_.begin();i!=times_.end();++i)
    ++buckets[static_cast<int>((*i)/period)-b0];

  for (int b=0;b<nb;++b) {
    out << datefromtime64((b+b0)*period) << ": ";
    out << std::setw(4) << std::setfill(' ') << buckets[b] << std::endl;
  }
}
